import inspect
import logging
import threading
from typing import Callable, Collection

from .types import (
    BaseAsyncEvent,
    BaseEvent,
    BaseEventChannel,
    BaseThreadedEvent,
)

logger = logging.getLogger("utils.events")


class Event(BaseEvent):

    def __init__(self, name: str | None = None):
        self._name: str | None = name
        self._channel: type[BaseEventChannel] | None = None
        self._subscribers: list[Callable] = []
        self._lock = threading.RLock()

        if name:
            logger.info(f"Event registered: {name}")

    @property
    def name(self) -> str:
        if not self._name:
            raise ValueError("Event must have a name")

        return self._name

    @property
    def channel(self) -> type[BaseEventChannel] | None:
        return self._channel

    @property
    def subscribers(self) -> Collection[Callable]:
        return tuple(self._subscribers)

    def subscribe(self, callback: Callable) -> Callable:
        if not callable(callback):
            raise TypeError("Callback must be a callable object")

        with self._lock:
            if not self.is_subscriber(callback=callback):
                self._subscribers.append(callback)
                logger.info(f"{callback} subscribed to {self}")

        return callback

    def unsubscribe(self, callback: Callable) -> None:
        with self._lock:
            if self.is_subscriber(callback=callback):
                self._subscribers.remove(callback)
                logger.info(f"{callback} unsubscribed from {self}")

    def is_subscriber(self, callback: Callable) -> bool:
        return callback in self._subscribers

    def publish(self, *args, **kwargs) -> None:
        logger.info(f"Publishing event: {self}")

        for callback in self._subscribers:
            callback(*args, **kwargs)

    def __set_name__(self, owner: type[BaseEventChannel], name: str):
        self._channel = owner

        if not self._name:
            self._name = name
            logger.info(f"Event registered: {self}")

    def __repr__(self):
        prefix = f"{self.channel}." if self.channel else ""
        return f"{prefix}{type(self).__name__}(name={self.name})"


class AsyncEvent(BaseAsyncEvent, Event):
    def __init__(self, name: str | None = None, nonblocking: bool = False):
        super().__init__(name=name)
        self._nonblocking = nonblocking

    async def publish_async(self, *args, **kwargs) -> None:
        logger.info(f"Publishing event: {self}")

        for callback in self._subscribers:
            if inspect.iscoroutinefunction(callback):
                await callback(*args, **kwargs)
            elif not self._nonblocking:
                callback(*args, **kwargs)


class ThreadedEvent(BaseThreadedEvent, Event):
    def __init__(self, name: str | None = None, daemons: bool = False):
        super().__init__(name=name)
        self._daemons = daemons

    def publish_threaded(self, *args, **kwargs) -> None:
        logger.info(f"Publishing event: {self}")

        for callback in self._subscribers:
            threading.Thread(
                target=callback,
                args=args,
                kwargs=kwargs,
                daemon=self._daemons,
            ).start()



from typing import Callable, Collection, Iterable

from .types import BaseEvent, BaseEventChannel, BaseEventChannelMetaclass


class EventChannelMetaclass(BaseEventChannelMetaclass):
    def __new__(mcs, name: str, bases: tuple, namespace: dict):
        cls = super().__new__(mcs, name, bases, namespace)
        cls.__events__ = {
            event.name: event
            for event in namespace.values()
            if isinstance(event, BaseEvent)
        }
        return cls

    @property
    def events(self) -> Collection[BaseEvent]:
        return self.__events__.values()

    def __contains__(self, event: str) -> bool:
        return event in self.__events__

    def __iter__(self):
        yield from self.events

    def __getitem__(self, event: str) -> "BaseEvent":
        if event not in self:
            raise KeyError(f"Event {event} not found at {self}")

        return self.__events__[event]

    def __repr__(self):
        return f"{self.__name__}"


class EventChannel(BaseEventChannel, metaclass=EventChannelMetaclass):
    pass



from abc import ABC, ABCMeta, abstractmethod
from typing import Callable, Collection


class BaseEventChannelMetaclass(ABCMeta):
    __events__: dict[str, "BaseEvent"]

    @property
    @abstractmethod
    def events(cls) -> Collection["BaseEvent"]: ...

    @abstractmethod
    def __contains__(self, event: str) -> bool: ...

    @abstractmethod
    def __repr__(cls): ...

    @abstractmethod
    def __getitem__(self, event: str) -> "BaseEvent": ...

    @abstractmethod
    def __iter__(self): ...


class BaseEventChannel(metaclass=BaseEventChannelMetaclass):
    pass


class BaseEvent(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def channel(self) -> type[BaseEventChannel] | None: ...

    @property
    @abstractmethod
    def subscribers(self) -> Collection[Callable]: ...

    @abstractmethod
    def subscribe(self, callback: Callable) -> Callable: ...

    @abstractmethod
    def unsubscribe(self, callback: Callable) -> None: ...

    @abstractmethod
    def is_subscriber(self, callback: Callable) -> bool: ...

    @abstractmethod
    def publish(self, *args, **kwargs) -> None: ...

    @abstractmethod
    def __repr__(self): ...


class BaseAsyncEvent(BaseEvent):
    @abstractmethod
    async def publish_async(self, *args, **kwargs) -> None: ...


class BaseThreadedEvent(BaseEvent):
    @abstractmethod
    def publish_threaded(self, *args, **kwargs) -> None: ...



from .channels import EventChannel
from .events import AsyncEvent, Event, ThreadedEvent
from .types import (
    BaseAsyncEvent,
    BaseEvent,
    BaseEventChannel,
    BaseThreadedEvent,
)

__all__ = [
    "BaseEventChannel",
    "BaseEvent",
    "BaseAsyncEvent",
    "BaseThreadedEvent",
    "Event",
    "AsyncEvent",
    "ThreadedEvent",
    "EventChannel",
]
